# Generated by Django 5.2 on 2025-06-01 12:26
from __future__ import annotations

from django.db import migrations, models

from undine.persisted_documents.validators import validate_document, validate_document_id


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="PersistedDocument",
            fields=[
                (
                    "document_id",
                    models.CharField(
                        max_length=255,
                        primary_key=True,
                        serialize=False,
                        validators=[validate_document_id],
                    ),
                ),
                ("document", models.TextField(validators=[validate_document])),
                ("created_at", models.DateTimeField(auto_now_add=True)),
            ],
            options={
                "verbose_name": "persisted document",
                "verbose_name_plural": "persisted documents",
                "abstract": False,
                "swappable": "UNDINE_PERSISTED_DOCUMENTS_MODEL",
            },
        ),
    ]
